=encoding utf8

This file is automatically generated.  DON'T EDIT THIS FILE MANUALLY.

=cut

package Unicode::LineBreak;

use constant {
    M => 4,
    D => 3,
    I => 2,
    P => 1,
};

our @LB_CLASSES = qw(LB_BK LB_CR LB_LF LB_NL LB_SP LB_OP LB_CL LB_QU LB_GL LB_NS LB_EX LB_SY LB_IS LB_PR LB_PO LB_NU LB_AL LB_ID LB_IN LB_HY LB_BA LB_BB LB_B2 LB_CB LB_ZW LB_CM LB_WJ LB_H2 LB_H3 LB_JL LB_JV LB_JT LB_SG LB_AI LB_SA LB_XX);

use constant {
    LB_BK => 0,
    LB_CR => 1,
    LB_LF => 2,
    LB_NL => 3,
    LB_SP => 4,
    LB_OP => 5,
    LB_CL => 6,
    LB_QU => 7,
    LB_GL => 8,
    LB_NS => 9,
    LB_EX => 10,
    LB_SY => 11,
    LB_IS => 12,
    LB_PR => 13,
    LB_PO => 14,
    LB_NU => 15,
    LB_AL => 16,
    LB_ID => 17,
    LB_IN => 18,
    LB_HY => 19,
    LB_BA => 20,
    LB_BB => 21,
    LB_B2 => 22,
    LB_CB => 23,
    LB_ZW => 24,
    LB_CM => 25,
    LB_WJ => 26,
    LB_H2 => 27,
    LB_H3 => 28,
    LB_JL => 29,
    LB_JV => 30,
    LB_JT => 31,
    LB_SG => 32,
    LB_AI => 33,
    LB_SA => 34,
    LB_XX => 35,
    LB_SAal => 36,
    LB_SAcm => 37,
    LB_NSidIter => 38,
    LB_NSidMasu => 39,
    LB_NSidKana => 40,
    LB_NSidLong => 41,
};

use constant {
    EA_Z => 0,
    EA_Na => 1,
    EA_N => 2,
    EA_A => 3,
    EA_W => 4,
    EA_H => 5,
    EA_F => 6,
    EA_NZ => 7,
    EA_AZ => 8,
    EA_AnLat => 9,
    EA_AnGre => 10,
    EA_AnCyr => 11,
    EA_WZ => 12,
};

1;
