#-*- perl -*-

package Unicode::LineBreak;

=encoding utf8

This file is automatically generated.  DON'T EDIT THIS FILE MANUALLY.

=cut

our $VERSION = '0.005.510';
our $UNICODE_VERSION = '5.1.0';

1;
